import time
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

# import dash
from dash_extensions.enrich import (
    Output,
    DashProxy,
    Input,
    MultiplexerTransform,
    html,
    State,
    ServersideOutput,
    ServersideOutputTransform,
)

# from dash_bootstrap_components import themes
# import dash_mantine_components as dmc
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from dash import MATCH, ALL, dcc, no_update
from flask_caching import Cache
from plotly_resampler import FigureResampler
from trace_updater import TraceUpdater
import uuid

# import pandas as pd


def generate_raw_signal(length=2_000_000):
    # Data that will be used for the plotly-resampler figures, serialized
    x = np.arange(length)
    noisy_sin = (3 + np.sin(x / 200) + np.random.randn(len(x)) / 10) * x / 1_000
    return noisy_sin
    # return pd.Series(noisy_sin).to_json(orient='values')


def y_parsing_func(data, expansion_factor) -> np.ndarray:
    "Dummy function which can be replaced with more advanced signal processing logic."
    return data * expansion_factor ** np.sqrt(np.arange(len(data)))


def y_parsing_func2(data, expansion_factor) -> np.ndarray:
    "Dummy function which can be replaced with more advanced signal processing logic."
    return (
        data * expansion_factor ** np.sqrt(np.arange(len(data)))
        + np.random.randn() * 1000
    )


def generate_plot_div(id, uid):

    children = [
        # Inputs
        dbc.Row(
            [
                dbc.Col(html.H3("Expansion factor:", style={"textAlign": "right"})),
                dbc.Col(
                    dbc.Input(
                        id={"type": "expansion-factor", "id": id, "index": uid},
                        **dict(min=0.5, max=1.02, step=0.005, value=1, type="number"),
                    )
                ),
                # dbc.Col(dbc.Button("plot chart", id={'type':'"plot-button"', 'id':id})),
            ],
            justify="center",
            align="center",
        ),
        # Graph
        dcc.Graph(
            id={"type": "graph", "id": id, "index": uid},
            animate=True,
            figure=go.Figure(),
            config={
                "showAxisDragHandles": True,
                "showAxisRangeEntryBoxes": True,
                "toImageButtonOptions": {
                    "format": "svg",  # one of png, svg, jpeg, webp
                    "filename": "plot",
                },
            },
            style={"width": "100%", "mx": 50, "minHeight": "300px"},
        ),
        # Aux stores
        # store used to link the server side cached FigureResampler object
        dcc.Loading(
            dcc.Store(
                id={"type": "store-graph", "id": id, "index": uid},
                storage_type="session",
            )
        ),
        # Interval to update the graph after a callback, that processes the data
        # after an input parameter change, is completed
        # dcc.Interval(
        #     id={"type": "interval", "id":id, "index": uid}, max_intervals=0,
        #     interval=100 # OBS1: It needs to be greater than the time it takes to store the data in the cache
        # ),
        # dcc.Interval(
        #     id={"type": "interval-relayout", "id":id, "index": uid}, max_intervals=0,
        #     interval=1
        # ),
        TraceUpdater(id={"type": "trace-updater", "id": id, "index": uid}, gdID=uid),
    ]

    return children


def generate_plot(data, n_plots, ylabels, var_names=None):

    fig = FigureResampler(
        make_subplots(rows=n_plots, cols=1, shared_xaxes=True, vertical_spacing=0.1)
    )

    fig.update_layout(
        margin={"t": 50, "b": 50, "l": 30, "r": 30},
        height=600,
        legend=dict(
            yanchor="bottom",
            xanchor="right",
            x=0.99,
            y=0.01,
            orientation="h",
        ),
        dragmode="zoom",
        newshape_line_color="#00372B",
        modebar_add=[
            "drawline",
            "drawopenpath",
            "drawcircle",
            "drawrect",
            "eraseshape",
        ],
    )

    fig["layout"]["yaxis"]["title"] = ylabels[0]
    for plt_idx in range(n_plots):
        # Set y-axes title
        if plt_idx > 0:
            fig["layout"][f"yaxis{plt_idx+1}"]["title"] = ylabels[plt_idx]

        # Add traces
        if plt_idx < len(data):
            # Multiple traces in the same subplot
            if isinstance(data[plt_idx], list):
                for trace_idx in range(len(data[plt_idx])):
                    # x = data[plt_idx][trace_idx].index.values
                    # y = data[plt_idx][trace_idx].values

                    x = np.arange(len(data[plt_idx][trace_idx]))
                    y = data[plt_idx][trace_idx]

                    fig.add_trace(
                        go.Scattergl(
                            x=[],
                            y=[],
                            name=var_names[plt_idx][trace_idx]
                            if var_names is not None
                            else None,
                        ),
                        hf_x=x,
                        hf_y=y,
                        row=plt_idx + 1,
                        col=1,
                    )

            else:  # Single trace in the subplot
                # x = data[plt_idx].index.values
                # y = data[plt_idx].values

                x = np.arange(len(data[plt_idx]))
                y = data[plt_idx]

                fig.add_trace(
                    go.Scattergl(
                        x=[],
                        y=[],
                        name=var_names[plt_idx] if var_names is not None else None,
                    ),
                    hf_x=x,
                    hf_y=y,
                    row=plt_idx + 1,
                    col=1,
                )

    return fig


app = DashProxy(
    __name__,
    transforms=[ServersideOutputTransform(), MultiplexerTransform()],
    external_stylesheets=[dbc.themes.BOOTSTRAP],
    title="Update traces test",
    update_title=False,
)


def generate_layout():
    uid = str(uuid.uuid4())

    return html.Div(
        [
            # Session id store
            dcc.Store(data=str(uuid.uuid4()), id="session-id"),
            # Processed signal data store and params pd.Series( test_data['TT-AQU-107a'] ).to_json(orient='values')
            dcc.Store(id={"type": "signal_data", "index": uid}),
            # Header
            html.H1(
                "Plotly-Resampler: modify trace data", style={"textAlign": "center"}
            ),
            html.Hr(),
            # Content. Here only a single element is created, but it could be extended
            html.Div(
                id="content-container", children=generate_plot_div("step1", uid=uid)
            ),
        ]
    )


cache = Cache(
    app.server,
    config={
        "CACHE_TYPE": "filesystem",
        "CACHE_DIR": "cache-directory",
        # should be equal to maximum number of users on the app at a single time
        # higher numbers will store more data in the filesystem / redis cache
        "CACHE_THRESHOLD": 20,
    },
)

app.layout = generate_layout()

# ------------------------------------ DASH logic -------------------------------------

# Input changes ---> Callback to process data ---> Callback to update the graph
# Relayout in front end OR Callback to update the graph ---> Callback to refresh graph in front end

# Callback to process data
@app.callback(
    # Output({"type": "interval", "id":"step1", "index": MATCH}, "n_intervals"),
    ServersideOutput({"type": "signal_data", "index": MATCH}, "data"),
    Input({"type": "expansion-factor", "id": "step1", "index": MATCH}, "value"),
    State({"type": "signal_data", "index": MATCH}, "data"),
)
def process_signal(value, signal_data):
    print('process_signal')
    if value is None:
        print('\tvalue is None')
        raise PreventUpdate()

    # Initialize data
    if not signal_data:
        raw_data = generate_raw_signal()
        signal_data = {"raw_data": raw_data}
    else:
        # raw_data = pd.read_json(signal_data['raw_data'], typ='series')
        raw_data = signal_data["raw_data"]

    processed_signal1 = y_parsing_func(raw_data, value)
    processed_signal2 = y_parsing_func2(raw_data, value)

    signal_data.update(
        {"processed_signal1": processed_signal1, "processed_signal2": processed_signal2}
    )
    return signal_data


# Callback to update the graph after data is processed
@app.callback(
    [
        Output({"type": "graph", "id": "step1", "index": MATCH}, "figure"),
        ServersideOutput(
            {"type": "store-graph", "id": "step1", "index": MATCH}, "data"
        ),
    ],
    # A change in the signal data store, triggers the update of the graph
    Input({"type": "signal_data", "index": MATCH}, "data"),
    [
        State({"type": "store-graph", "id": "step1", "index": MATCH}, "data"),
        State({"type": "graph", "id": "step1", "index": MATCH}, "relayoutData"),
    ],
    prevent_initial_call=True,
)
def process_and_update_figure(signal_data, fig, relayout):
    if signal_data is None:
        raise PreventUpdate()
    if signal_data.get("processed_signal1", None) is None:
        raise PreventUpdate()

    # Retrieve data to plot from the store
    raw_data = signal_data["raw_data"]
    processed_signal1 = signal_data["processed_signal1"]
    processed_signal2 = signal_data["processed_signal2"]
    print('signal data changed, lets update the graph')
    print('processed_signal1', processed_signal1)
    print('processed_signal2', processed_signal2)

    if fig is None:
        # Figure does not exist, create it
        data = [raw_data, [processed_signal1, processed_signal2]]
        n_plots = 2
        ylabels = ["Original signal", "Processed signals"]
        var_names = ["original signal", ["processed signal 1", "processed signal 2"]]

        fig = generate_plot(
            data=data, n_plots=n_plots, ylabels=ylabels, var_names=var_names
        )
    else:
        # 0. Update the lower trace raw data via the `hf_data` property
        # OBS2: Need to update every trace otherwise construct_update_data fails
        fig.hf_data[0]["y"] = raw_data
        fig.hf_data[1]["y"] = processed_signal1
        fig.hf_data[2]["y"] = processed_signal2

        # NOTE: As we send back the `dcc.Figure` and not the `updateData`, we will
        # construct the `updateData` based on the `relayout` and use it to set the
        # `dcc.Figure` its trace data accordingly

        # 1. Perform a relayout on the figure based on the front-end layout
        #    So that the figure range of the to be returned figure will be correct
        fig.plotly_relayout(relayout)

        # 2. Alter the relayout dict to make sure that `construct_update_data` will
        #    trigger (and thus not return `dash.no_update`)
        for ax in fig.layout.to_plotly_json().keys():
            if not ax.startswith("xaxis"):
                continue

            # Current axis relayout
            if f"{ax}.range[0]" in relayout:
                print('current relayout')
                relayout.update({f"{ax}.range[0]":relayout[f"{ax}.range[0]"]-1, 
                                 f"{ax}.range[1]":relayout[f"{ax}.range[1]"]+1})
            
            # Past axis relayout
            elif f"{ax}.range" in relayout:
                print('sustained relayout')
                relayout.update({f"{ax}.range[0]":relayout[f"{ax}.range"][0]-1, 
                                 f"{ax}.range[1]": relayout[f"{ax}.range"][1]+1})
                relayout.pop(f"{ax}.range")

            # No axis relayouts
            else:
                print('no relayouts')
                axes = [trace['xaxis'] for trace in fig.data]
                if ax[-1].isnumeric():
                    trace_idx=axes.index(f'x{ax[-1]}')
                else:
                    trace_idx=axes.index('x')
                    relayout = {}
                
                relayout.update({f"{ax}.range[0]":fig.hf_data[ trace_idx ]["x"][0]-1,
                                 f"{ax}.range[1]":fig.hf_data[ trace_idx ]["x"][-1]+1})

        # 3. Construct the updateData
        #    and update the `dcc.Figure` data with the updated data
        for data in fig.construct_update_data(relayout)[1:]:
            trace_idx = data.pop("index")  # pop returns the deleted key value
            # data['uid'] = fig.data[trace_idx]['uid']
            fig.data[trace_idx].update(data, overwrite=True)
        fig.plotly_relayout(relayout)
        print('fig updated')
    return fig, fig


# Callback to update the graph after relayouts

# To update after relayouts, it cannot be a callback with MATCH id, it has to be a callback for each graph
# If not, the figure imported from the store-graph is a string for some reason
@app.callback(
    Output({"type": "trace-updater", "id": "step1", "index": MATCH}, "updateData"),
    Input({"type": "graph", "id": "step1", "index": MATCH}, "relayoutData"),
    State({"type": "store-graph", "id": "step1", "index": MATCH}, "data"),
    prevent_initial_call=True,
)
def update_fig(relayoutdata, fig):
    time.sleep(0.5)
    if fig is not None:
        return fig.construct_update_data(relayoutdata)
    return no_update


# --------------------------------- Running the app ---------------------------------
if __name__ == "__main__":
    app.run_server(debug=True, port=9023)
